/*
 * wait.c
 *
 * Copyright (C) 2006  Insigme Co., Ltd
 *
 * This software has been developed while working on the Linux Unified Kernel
 * project (http://www.longene.org) in the Insigma Research Institute,  
 * which is a subdivision of Insigma Co., Ltd (http://www.insigma.com.cn).
 * 
 * The project is sponsored by Insigma Co., Ltd.
 *
 * The authors can be reached at linux@insigma.com.cn.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of  the GNU General  Public License as published by the
 * Free Software Foundation; either version 2 of the  License, or (at your
 * option) any later version.
 *
 * Revision History:
 *   Jan 2006 - Created.
 */
 
/*
 * wait.c: thread wait on object 
 * Refered to Kernel-win32 code
 */
#include <linux/mm.h>
#include <linux/delay.h>
#include <asm/uaccess.h>
#include <apc.h>
#include <file.h>
#include <io.h>
#include <mutex.h>
#include <object.h>
#include <pefile.h>
#include <unistr.h>
#include <attach.h>
#include <event.h>
#include <handle.h>
#include <ke.h>
#include <ntstatus.h>
#include <objwait.h>
#include <process.h>
#include <section.h>
#include <semaphore.h>
#include <virtual.h>
#include <win32.h>
#include <thread.h>
#include <w32syscall.h>

#ifdef CONFIG_UNIFIED_KERNEL

/* Tells us if the Timer or Event is a Syncronization or Notification Object */
#define TIMER_OR_EVENT_TYPE 0x7L

/* One of the Reserved Wait Blocks, this one is for the Thread's Timer */
#define TIMER_WAIT_BLOCK 0x3L

void call_kapc(void){
	/* FIXME */
}
EXPORT_SYMBOL(call_kapc);

VOID
STDCALL
block_thread(PNTSTATUS Status,
		UCHAR Alertable,
		ULONG WaitMode,
		UCHAR WaitReason,
		PULONG_PTR Timeout)
{
	struct kthread * thread = (struct kthread *)get_current_ethread();
	struct kwait_block * wait_block;

	ktrace("block thread\n");
	if (thread->apc_state.kapc_pending) {
		/* Remove Waits */
		wait_block = thread->wait_block_list;
		do {
			list_del(&wait_block->wait_list_entry);
			wait_block = wait_block->next_wait_block;
		} while (wait_block!= thread->wait_block_list);
		thread->wait_block_list = NULL;

		/* Ready Dispatch it and return status */
		call_kapc();

		if (Status != NULL) *Status = STATUS_KERNEL_APC;

	} else {
		/* Set the Thread Data as Requested */
		thread->alertable = Alertable;
		thread->wait_mode = (UCHAR)WaitMode;
		thread->wait_reason = WaitReason;

		local_irq_enable();
		set_current_state(TASK_INTERRUPTIBLE);
		*Timeout = schedule_timeout(*Timeout);
		local_irq_disable();

		/* Dispatch it and return status */
		if (signal_pending(current)) {
			/* thread is signaled
			 * Remove Waits */
			wait_block = thread->wait_block_list;
			do {
				list_del(&wait_block->wait_list_entry);
				wait_block = wait_block->next_wait_block;
			} while (wait_block!= thread->wait_block_list);
			thread->wait_block_list = NULL;
			*Status = -EINTR; /* Linux error code for special check */
			return;
		}

		if(Status) {
			if (!(*Timeout)){ 
				/* thread is signaled
				 * Remove Waits */
				wait_block = thread->wait_block_list;
				do {
					list_del(&wait_block->wait_list_entry);
					wait_block = wait_block->next_wait_block;
				} while (wait_block!= thread->wait_block_list);
				thread->wait_block_list = NULL;
				*Status = STATUS_TIMEOUT;
			} else *Status = ((struct kthread * )current->ethread)->wait_status;
		}
	}

	/* FIXME */
#if 0
	KfLowerIrql(Thread->WaitIrql);
#endif
}
EXPORT_SYMBOL(block_thread);

VOID
inline
check_alert(BOOLEAN Alertable,
		struct kthread *CurrentThread,
		KPROCESSOR_MODE WaitMode,
		PNTSTATUS Status)
{
	/* At this point, we have to do a wait, so make sure we can make the thread Alertable if requested */
	if (Alertable) {

		/* If the Thread is Alerted, set the Wait Status accordingly */
		if (CurrentThread->alerted[(int)WaitMode]) {
			CurrentThread->alerted[(int)WaitMode] = false;
			*Status = STATUS_ALERTED;

			/* If there are User APCs Pending, then we can't really be alertable */
		} else if ((!list_empty(&CurrentThread->apc_state.apc_list_head[UserMode])) &&
				(WaitMode == UserMode)) {
			CurrentThread->apc_state.uapc_pending= true;
			*Status = STATUS_USER_APC;
		}

		/* If there are User APCs Pending and we are waiting in usermode, then we must notify the caller */
	} else if ((CurrentThread->apc_state.uapc_pending) && (WaitMode == UserMode)) {
		*Status = STATUS_USER_APC;
	}

	*Status = STATUS_WAIT_0;
}
EXPORT_SYMBOL(check_alert);


VOID
satisfy_object_wait(struct dispatcher_header * Object,
		struct kthread *Thread)
{ 
	/* Special case for Mutants */ 
	if (Object->type == MutantObject) { 
		Object->signal_state--; 

		/* Check if it's now non-signaled */ 
		if (Object->signal_state == 0) 
		{ 
			((struct kmutant *)Object)->owner_thread = Thread; 
			Thread->kernel_apc_disable -= ((struct kmutant *)Object)->apc_disable; 

			/* Check if it's abandoned */ 
			if (((struct kmutant *)Object)->abandoned) { 
				/* Unabandon it */ 
				((struct kmutant *)Object)->abandoned = FALSE; 
				Thread->wait_status = STATUS_ABANDONED; 
			} 

			list_add(&((struct kmutant *)Object)->mutant_list_entry, &Thread->mutant_list_head); 
		} 
	} else if ((Object->type & TIMER_OR_EVENT_TYPE) == EventSynchronizationObject) { 
		Object->signal_state= 0; 
	} else if (Object->type == SemaphoreObject) { 
		Object->signal_state--; 
	} 
}

void
satisfy_multi_obj_waits(struct kwait_block * WaitBlock)
{ 
	struct kwait_block * first_block = WaitBlock; 
	struct kthread * wait_thread = WaitBlock->thread; 

	/* Loop through all the Wait Blocks, and wake each Object */ 
	do { 
		/* Wake the Object */ 
		satisfy_object_wait(WaitBlock->object, wait_thread); 
		WaitBlock = WaitBlock->next_wait_block; 
	} while (WaitBlock != first_block);
}

NTSTATUS
STDCALL
wait_for_single_object(PVOID Object,
		KWAIT_REASON WaitReason,
		KPROCESSOR_MODE WaitMode,
		BOOLEAN Alertable,
		PLARGE_INTEGER Timeout)
{
	NTSTATUS status;
	struct kwait_block wait_block[MAXIMUM_WAIT_OBJECTS];
	
	status = wait_for_multi_objs(1,
				&Object,
				WaitAll,
				UserRequest,
				WaitMode,
				Alertable,
				Timeout,
				wait_block);
	return status;
}
EXPORT_SYMBOL(wait_for_single_object);

BOOL
is_object_waitable(PVOID Object)
{
	return 0;
}


VOID
wait_test(struct dispatcher_header * Object,
		KPRIORITY Increment)
{
	struct list_head * wait_entry; 
	struct list_head * wait_list; 
	struct kwait_block * cur_wait_block; 
	struct kwait_block * next_wait_block; 
	struct kthread * wait_thread; 

	wait_list = &Object->wait_list_head; /* TODO kernel */
	wait_entry = wait_list->next; 
	while ((wait_entry != wait_list) && (Object->signal_state > 0)) { 
		/* Get the current wait block */ 
		cur_wait_block = list_entry(wait_entry, struct kwait_block, wait_list_entry); 
		wait_thread = cur_wait_block->thread; 

		if (cur_wait_block->wait_type == WaitAny) { 

			wait_entry = wait_entry->prev; 
			satisfy_object_wait(Object, wait_thread); 
		} else { 

			next_wait_block = cur_wait_block->next_wait_block; 

			/* Loop first to make sure they are valid */ 
			while (next_wait_block != cur_wait_block) { 
				if (!is_object_signaled(next_wait_block->object, wait_thread)) { 
					/* It's not, move to the next one */ 
					goto SkipUnwait; 
				} 

				next_wait_block = next_wait_block->next_wait_block;
			} 

			/* All the objects are signaled, we can satisfy */ 
			wait_entry = wait_entry->prev; 
			satisfy_multi_obj_waits(cur_wait_block); 
		} 

		/* All waits satisfied, unwait the thread */ 
		abort_wait_thread(wait_thread, cur_wait_block->wait_key, Increment); 

SkipUnwait: 
		wait_entry = wait_entry->next; 
	} 

}
EXPORT_SYMBOL(wait_test);

NTSTATUS 
STDCALL
wait_for_multi_objs(ULONG Count,
		PVOID Object[],
		WAIT_TYPE WaitType,
		KWAIT_REASON WaitReason,
		KPROCESSOR_MODE WaitMode,
		BOOLEAN Alertable,
		PLARGE_INTEGER Timeout,
		struct kwait_block *WaitBlockArray)
{
	struct dispatcher_header * cur_obj;
	struct kwait_block * wait_block;
	struct kthread * cur_thread = (struct kthread *)get_current_ethread();
	unsigned long all_objects_signaled;
	unsigned long wait_index;
	NTSTATUS status, wait_status;
	struct timespec ts;
	long timeout;

	if (Timeout) {
		s32 rem;
		ts.tv_sec = div_s64_rem(-Timeout->QuadPart * 100, 1000000000, &rem);
		ts.tv_nsec = rem;
		timeout = timespec_to_jiffies(&ts) + (ts.tv_sec || ts.tv_nsec);
	}
	else
		timeout = MAX_SCHEDULE_TIMEOUT;

	/* Check if the lock is already held */
	if (cur_thread->wait_next)  {
		/* Lock is held, disable Wait Next */
		cur_thread->wait_next = false;
	} else {
		local_irq_disable();
		/* FIXME Lock not held, acquire it */
	}

	/* Make sure the Wait Count is valid for the Thread and Maximum Wait Objects */
	if (!WaitBlockArray) {
		/* FIXME Debug Check in regards to the Thread Object Limit */

		/* Use the Thread's Wait Block */
		WaitBlockArray = &cur_thread->wait_block[0];
	} else {
		/* FIXME Using our own Block Array. Check in regards to System Object Limit */
	}

	/* Start the actual Loop */
	do {
		wait_status = cur_thread->wait_status;
		cur_thread->wait_block_list = wait_block = WaitBlockArray;
		all_objects_signaled = true;

		/* First, we'll try to satisfy the wait directly */
		for (wait_index = 0; wait_index < Count; wait_index++) {
			cur_obj = (struct dispatcher_header * )Object[wait_index];

			if (cur_obj->type == IO_TYPE_FILE) {
				cur_obj = (struct dispatcher_header *)(&((PFILE_OBJECT)cur_obj)->Event);
			}

			if (is_object_signaled(cur_obj, cur_thread)) {
				if (WaitType == WaitAny) {
					if (cur_obj->signal_state!= (LONG)MINLONG) {
						/* It has a normal signal state, so unwait it and return */
						satisfy_object_wait(cur_obj, cur_thread);
						if(STATUS_ABANDONED == cur_thread->wait_status)
							status = cur_thread->wait_status = (STATUS_ABANDONED | wait_index);
						else
							status = STATUS_WAIT_0 | wait_index;
						goto WaitDone;

					} else {
						/* Is this a Mutant? */
						if (cur_obj->type == MutantObject) {
							/* FIXME exception*/
						}
					}
				}

			} else {
				/* One of the objects isn't signaled... if this is a WaitAll, we will fail later */
				all_objects_signaled = false;
			}

			/* Set up a Wait Block for this Object */
			wait_block->object = cur_obj;
			wait_block->thread = cur_thread;
			if(WaitAny == WaitType)
				wait_block->wait_key = (USHORT)(STATUS_WAIT_0 + wait_index);
			else  
				wait_block->wait_key = (USHORT)(STATUS_WAIT_0);
			wait_block->wait_type = (USHORT)WaitType;
			wait_block->next_wait_block = wait_block + 1;

			wait_block = wait_block->next_wait_block;
		}

		/* Return to the Root Wait Block */
		wait_block--;
		wait_block->next_wait_block = WaitBlockArray;

		if ((WaitType == WaitAll) && (all_objects_signaled)) {
			wait_block = cur_thread->wait_block_list;

			satisfy_multi_obj_waits(wait_block);
			if(STATUS_ABANDONED == cur_thread->wait_status)
				status = STATUS_ABANDONED_WAIT_0;
			else status = STATUS_WAIT_0;
			goto WaitDone;
		}

		/* Make sure we can satisfy the Alertable request */
		check_alert(Alertable, cur_thread, WaitMode, &status);

		cur_thread->wait_status = status;
		wait_block = cur_thread->wait_block_list;

		do {
			cur_obj = wait_block->object;
			list_add(&wait_block->wait_list_entry, &cur_obj->wait_list_head);
			wait_block = wait_block->next_wait_block;
		} while (wait_block != WaitBlockArray);

		/* Handle Kernel Queues */

		block_thread(&status,
				Alertable,
				WaitMode,
				(UCHAR)WaitReason,
				&timeout);

		/* Check if we were executing an APC */
		if (status != STATUS_KERNEL_APC) {
			local_irq_enable();
			return status;
		}

	} while (true);

WaitDone:
	local_irq_enable();
	/* Release the Lock, we are done */
	if (Timeout) {
		jiffies_to_timespec(timeout, &ts);
		Timeout->QuadPart = -(ts.tv_sec * 10000000L + ts.tv_nsec / 100);
	}

	return status;
}
EXPORT_SYMBOL(wait_for_multi_objs);

/* Must be called with the dispatcher lock held */
VOID
abort_wait_thread(struct kthread *Thread,
		NTSTATUS WaitStatus,
		KPRIORITY Increment)
{
	struct kwait_block *  wait_block; 

	/* If we are blocked, we must be waiting on something also */ 
	if(!((((struct ethread *) Thread)->et_task->state & 
					(TASK_STOPPED | TASK_TRACED | TASK_INTERRUPTIBLE | TASK_UNINTERRUPTIBLE)) 
				&& Thread->wait_block_list != NULL))
		return;

	wait_block = Thread->wait_block_list; 
	do { 
		list_del(&wait_block->wait_list_entry); 
		wait_block = wait_block->next_wait_block; 
	} while (wait_block != Thread->wait_block_list); 

	if(STATUS_ABANDONED == Thread->wait_status) Thread->wait_status += WaitStatus - STATUS_WAIT_0;
	else Thread->wait_status = WaitStatus;

	/* FIXME : Check if there's a Thread Timer */ 

	/* Reschedule the Thread */ 
	set_tsk_need_resched(current);
	wake_up_process(((struct ethread *)Thread)->et_task);
}
EXPORT_SYMBOL(abort_wait_thread);
#endif

